import logging
import os
import re
import sys

import yaml

from .settings import CONFIG_FILE
from .utils import colour_text


log = logging.getLogger(__name__)


def normalize_key_value(key, value):
    if key.endswith('regex'):
        return key.replace('regex', 'pattern'), value and value.pattern

    return key, value


def dict_to_kwarg_str(dct):
    return ', '.join(
        f'{key}={value!r}' for key, value in (normalize_key_value(*kv) for kv in dct.items())
    )


class BadConfig(ValueError):
    pass


class ConfigObject:
    @staticmethod
    def normalize_path(path: str):
        path = os.path.normpath(path.strip())

        if os.path.isabs(path):
            cwd = os.getcwd()
            if os.path.commonpath([path, cwd]) != cwd:
                # this path is not in our directory
                return None

            return os.path.relpath(path)

        if path.split(os.sep)[0] == os.pardir:
            # this file is not in our directory
            return None

        return path

    def _check_and_init_pattern(self, attr, pattern):
        assert attr.endswith('pattern')

        if pattern is None:
            regex = None
        else:
            try:
                regex = re.compile(pattern, re.IGNORECASE)
            except re.error as exc:
                raise BadConfig(f'{attr} is not a valid regular expression: {exc}')

        setattr(self, attr.replace('pattern', 'regex'), regex)

    def __repr__(self):
        return f'{type(self).__name__}({dict_to_kwarg_str(self.__dict__)})'


class MonitorConfig(ConfigObject):
    # ----------------------------------------------------------------------------------------------
    # Instance API
    # ----------------------------------------------------------------------------------------------
    def __init__(
        self,
        name,
        command=None,
        file_pattern=None,
        problem_line_file_pattern=None,
        foreground_colour=None,
        background_colour=None,
    ):
        self.name = name

        if (
            not isinstance(command, list)
            or len(command) == 0
            or any(not isinstance(command_arg, str) for command_arg in command)
        ):
            raise BadConfig('command must be a non-empty list of the command and its arguments')

        self.command = command

        if not isinstance(file_pattern, (type(None), str)):
            raise BadConfig('file_pattern must be a valid regular expression')

        self._check_and_init_pattern('file_pattern', file_pattern)
        self._check_and_init_pattern('problem_line_file_pattern', problem_line_file_pattern)

        if foreground_colour is not None:
            try:
                colour_text('', foreground=foreground_colour)
            except KeyError:
                raise BadConfig('Unknown foreground color')
        self.foreground_colour = foreground_colour

        if background_colour is not None:
            try:
                colour_text('', background=background_colour)
            except KeyError:
                raise BadConfig('Unknown background color')
        self.background_colour = background_colour

    def includes_file(self, filepath):
        filedir, filename = os.path.split(filepath)
        return self.file_regex is None or bool(self.file_regex.search(filename))

    def extract_file_from_problem_line(self, line):
        if self.problem_line_file_regex is None:
            return None

        mo = self.problem_line_file_regex.search(line)
        if mo is None:
            return None

        try:
            return self.normalize_path(mo.group(1))
        except IndexError:
            return None

    def badge_for_number(self, number):
        if number == 0:
            return ''
        kwargs = {}
        if self.foreground_colour is not None:
            kwargs['foreground'] = self.foreground_colour
        if self.background_colour is not None:
            kwargs['background'] = self.background_colour
        return colour_text(f' {number} ', **kwargs)

        # ----------------------------------------------------------------------------------------------
        # Magic methods
        # ----------------------------------------------------------------------------------------------
        return self.name


class GlobalConfig(ConfigObject):
    def __init__(self, monitors=None):
        if monitors is None:
            raise BadConfig('No monitors specified')

        if not isinstance(monitors, dict):
            raise BadConfig(
                'monitors section must be a dictionary of named monitors (each itself a dictionary)'
            )

        if len(monitors) == 0:
            raise BadConfig('Empty monitors dictionary')

        self.monitors = {}
        self.monitor_errors = []
        for name, mon_con in monitors.items():

            try:
                self.monitors[name] = clean_monitor_config(name, mon_con)
            except BadConfig as exc:
                self.monitor_errors.append(f'Monitor "{name}" invalid: {exc}')

        if len(self.monitors) == 0:
            raise BadConfig(f'No valid monitors: {"; ".join(self.monitor_errors)}')

        log.debug('cleaned config: %r', self)


def clean_monitor_config(name, monitor_config):
    if not isinstance(monitor_config, dict):
        raise BadConfig('Not a dictionary')

    return MonitorConfig(name, **monitor_config)


def clean_config(config):
    if not isinstance(config, dict):
        raise BadConfig('Config is not a dictionary')

    return GlobalConfig(**config)


def load_config_file(config_file):
    log.debug('Load config')
    if not os.path.exists(config_file):
        raise BadConfig('No such config file')

    with open(config_file, "r") as stream:
        try:
            config = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            oneline_exc = str(exc).replace('\n', ' ')
            raise BadConfig(f'Invaid YAML file: {oneline_exc}') from exc

    return clean_config(config)


def load_config_or_exit():
    try:
        return load_config_file(CONFIG_FILE)
    except BadConfig as exc:
        log.error('Bad config file %s: %s', CONFIG_FILE, exc)
        sys.exit(1)



import os

CONFIG_FILE = 'lintmon.yaml'
DEFAULT_IGNORED_DIRECTORY_NAMES = [r'\.lintmon', r'\.git', 'node_modules', '__pycache__']
STATE_DIR = '.lintmon'
PID_FILE = os.path.join(STATE_DIR, 'pid')
STOP_WAIT_SECONDS = 10
STOP_FILE = os.path.join(STATE_DIR, 'stop')



def colour_text(text, foreground='white', background='black'):
    foreground_colours = {
        'black': '30',
        'red': '31',
        'green': '32',
        'yellow': '33',
        'blue': '34',
        'pink': '35',
        'cyan': '36',
        'grey': '37',
        'gray': '37',
        'white': '38',
    }
    background_colours = {
        'black': '40',
        'red': '41',
        'green': '42',
        'yellow': '43',
        'blue': '44',
        'pink': '45',
        'cyan': '46',
        'grey': '47',
        'gray': '47',
    }

    fg = foreground_colours[foreground]
    bg = background_colours[background]

    return f'\x1b[{fg}m\x1b[{bg}m{text}\x1b[0m'


def diff_problem_lines(lines_before, lines_after):
    lines_before_set = {(file, line) for file, lines in lines_before.items() for line in lines}
    lines_after_set = {(file, line) for file, lines in lines_after.items() for line in lines}
    before_not_after = lines_before_set - lines_after_set
    after_not_before = lines_after_set - lines_before_set
    return [
        (plus_minus, file, line)
        for plus_minus, set_ in [('-', before_not_after), ('+', after_not_before)]
        for file, line in set_
    ]


def first(iter_):
    for item in iter_:
        return item

    return None


def gb(iter_, func):
    dct = {}
    for thing in iter_:
        key = func(thing)
        if key is None:
            continue
        dct.setdefault(key, []).append(thing)
    return dct


def lf(func, iter_):
    return list(filter(func, iter_))



import logging
import logging.config
import os
import re
from signal import SIGTERM
from subprocess import DEVNULL, Popen
from time import sleep, time
from typing import Optional

import psutil

from .config import load_config_file, BadConfig, load_config_or_exit
from .lintmon import Lintmon
from .settings import (
    CONFIG_FILE,
    DEFAULT_IGNORED_DIRECTORY_NAMES,
    PID_FILE,
    STATE_DIR,
    STOP_FILE,
    STOP_WAIT_SECONDS,
)
from .utils import colour_text


LOG_LEVEL = os.environ.get('LOG_LEVEL') or 'WARNING'
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'simple': {'format': '%(levelname)s %(message)s'},
        'file': {'format': '%(asctime)s %(levelname)s: %(message)s'},
    },
    'handlers': {
        'console': {'level': LOG_LEVEL, 'class': 'logging.StreamHandler', 'formatter': 'simple'},
        'file': {
            'level': logging.INFO,
            'class': 'logging.FileHandler',
            'filename': '.lintmon/output.log',
            'formatter': 'file',
        },
        'null': {'class': 'logging.NullHandler'},
    },
    'root': {'level': 'DEBUG', 'handlers': ['console', 'file'], 'propagate': False},
}

try:
    logging.config.dictConfig(LOGGING)
except ValueError as exc:
    if "Unable to configure handler 'file'" not in str(exc):
        raise

log = logging.getLogger()


# --------------------------------------------------------------------------------------------------
# Entry points
# --------------------------------------------------------------------------------------------------
def lintmond():
    # if '--quiet' in sys.argv:
    #     root_logger = logging.getLogger()
    #     console_handler = first(h for h in root_logger.handlers if h.name == 'console')
    #     root_logger.removeHandler(console_handler)

    log.debug('lintmond main')

    lintmon = Lintmon(load_config_or_exit())
    lintmon.run()


def run_all():
    log.debug('Run all')
    config = load_config_or_exit()
    files = find_all_appropriate_files()
    lintmon = Lintmon(config)
    lintmon.update_sessions(files)

    print_sessions(lintmon.sessions)


def status():
    if not is_here():
        print('No lintmon.yaml in this directory')
        return

    if is_stopped():
        print('🛑 lintmon is stopped')
    else:
        if not lintmon_is_running():
            print('⚠️ lintmon is not running')
        else:
            print(f'✅ lintmon running pid {lintmon_pid()}')
    try:
        config = load_config_file(CONFIG_FILE)
    except BadConfig as exc:
        if 'No such config file' in str(exc):
            return 0
        print(colour_text(' ! ', background='red', foreground='white'), end='')
        return 1

    config = load_config_or_exit()

    lintmon = Lintmon(config)
    lintmon.load_latest_sessions()

    print_sessions(lintmon.sessions)


def status_prompt():
    if not is_here():
        return

    os.makedirs(STATE_DIR, exist_ok=True)

    if is_stopped():
        print(colour_text(' S ', background='red', foreground='white'), end='')
        return

    ensure_lintmon_is_running()

    try:
        config = load_config_file(CONFIG_FILE)
    except BadConfig as exc:
        if 'No such config file' in str(exc):
            return 0
        print(colour_text(' ! ', background='red', foreground='white'), end='')
        return 1

    config = load_config_or_exit()

    lintmon = Lintmon(config)
    lintmon.load_latest_sessions()

    print(lintmon.badges, end='')
    log.debug('status_prompt finished')


def start():
    if is_stopped():
        os.remove(STOP_FILE)

    ensure_lintmon_is_running()


def stop():
    if not is_stopped():
        with open(STOP_FILE, 'w'):
            pass

    pid = lintmon_pid()
    if pid is None:
        print('Not running')
        return

    print(f'Sending SIGTERM to {pid}')
    os.kill(pid, SIGTERM)

    print(f'Waiting up to {STOP_WAIT_SECONDS} for termination...')
    start = time()
    while time() < start + STOP_WAIT_SECONDS:
        if not lintmon_is_running():
            print('Terminated')
            return

        sleep(0.1)

    print(f'{pid} did not terminate')


# --------------------------------------------------------------------------------------------------
# Helpers
# --------------------------------------------------------------------------------------------------
def find_all_appropriate_files():
    full_file_paths = []
    for dirpath, subdirs, files in os.walk(os.curdir):
        for file in files:
            full_file_paths.append(os.path.join(dirpath, file))

        for ignored_subdir in DEFAULT_IGNORED_DIRECTORY_NAMES:
            try:
                subdirs.remove(ignored_subdir)
            except ValueError:
                pass

    return full_file_paths


def is_here():
    return os.path.exists(CONFIG_FILE)


def is_stopped():
    return os.path.exists(STOP_FILE)


def lintmon_pid() -> Optional[int]:
    if not os.path.exists(PID_FILE):
        log.debug('no pid file %s', PID_FILE)
        return None

    with open(PID_FILE) as pf:
        pid_line = pf.readline()

    mo = re.match(r'(\d+)', pid_line)
    if not mo:
        log.debug('Pid file does not match regex %s', pid_line)
        return None

    pid = int(mo.group(1))
    pid_alive = psutil.pid_exists(pid)

    if not pid_alive:
        log.debug('Pid %s is not alive', pid)
        os.remove(PID_FILE)
        return None
    return pid


def lintmon_is_running():
    return lintmon_pid() is not None


def print_sessions(sessions):
    for ms in sessions:
        if len(ms.problem_lines) == 0:
            print(f'✅ {ms} clean')
            continue

        coloured = ms.badge
        print(f'{coloured} {ms} output:')
        for ol in (pl for pls in ms.problem_lines.values() for pl in pls):
            print(f'  {ol}')


def run_lintmond():
    print('Starting lintmond...')
    proc = Popen(['lintmond', '--quiet'], stdin=DEVNULL, stderr=DEVNULL, stdout=DEVNULL)
    log.debug('lintmond running %s', proc.pid)

    with open(PID_FILE, 'w') as pf:
        pf.write(str(proc.pid))


def ensure_lintmon_is_running():
    if lintmon_is_running():
        log.info('lintmon running')
        return

    run_lintmond()



import logging
import os
from subprocess import Popen
from tempfile import TemporaryFile

from .settings import STATE_DIR
from .utils import diff_problem_lines, gb


log = logging.getLogger(__name__)


class MonitorSession:
    class States:
        initial = 'initial'
        running = 'running'
        complete = 'complete'

    @property
    def badge(self):
        num_problems = sum(len(lines) for lines in self.problem_lines.values())
        return self.config.badge_for_number(num_problems)

    @property
    def dirpath(self):
        return os.path.join(STATE_DIR, 'monitors', self.config.name)

    @property
    def problem_lines_filepath(self):
        return os.path.join(self.dirpath, 'problem_lines')

    def __init__(self, config, files):
        self.state = self.States.initial
        self.config = config
        self.files = files
        self.process = None
        self.output_file = None
        self.problem_lines = None

        self.initial_problem_lines = gb(
            self._read_lines_filepath(self.problem_lines_filepath),
            self.config.extract_file_from_problem_line,
        )
        log.debug('%s initial problem lines %s', self, self.initial_problem_lines)

    def start(self):
        assert self.state == self.States.initial

        if len(self.files) == 0:
            self.skip()
            return

        files = [f for f in self.files if os.path.exists(f)]

        if len(files) == 0:
            # all files were deleted, so mark as clear
            self.problem_lines = {f: [] for f in self.files}
            self.state = self.States.complete
            return

        self.state = self.States.running
        log.debug('Starting %s on %d files', self.config.command, len(files))
        self.output_file = TemporaryFile(mode='w+')
        try:
            self.process = Popen(
                [*self.config.command, *files], stdout=self.output_file, stderr=self.output_file,
            )
        except Exception as exc:
            self.problem_lines = {'.': [str(exc).replace('\n', ' ')]}
            self.process = None

    def join(self):
        if self.state == self.States.complete:
            return

        assert self.state == self.States.running
        if self.process is None:
            self.state = self.States.complete
            return

        self.process.communicate()
        self.output_file.flush()
        self.output_file.seek(0)
        olines = self.output_file.readlines()
        log.debug('%s output lines: %s', self, olines)
        self.problem_lines = gb(
            (line.strip() for line in olines), self.config.extract_file_from_problem_line,
        )
        # make sure we detect removal of problems by setting empty arrays for files that we
        # processed but didn't get output for
        for file in self.files:
            self.problem_lines.setdefault(file, [])
        self.output_file.close()
        self.output_file = None
        self.process = None
        self.state = self.States.complete

    def skip(self):
        # don't bother running, just use initial values as end values
        assert self.state == self.States.initial
        self.problem_lines = self.initial_problem_lines
        self.state = self.States.complete

    def save(self):
        # only able to save once it has been joined
        assert self.state == self.States.complete
        assert self.problem_lines is not None

        new_problem_lines_by_file = {
            **self.initial_problem_lines,
            **self.problem_lines,
        }

        log.debug('%s new_problem_lines_by_file %s', self, new_problem_lines_by_file)

        problem_line_diff = diff_problem_lines(
            self.initial_problem_lines, new_problem_lines_by_file
        )

        if len(problem_line_diff) == 0:
            log.debug('No change, no need to save %s', self)
            return

        log.info('Changes in %s:', self)
        for diff_entry in problem_line_diff:
            log.info('  %s %s', diff_entry[0], diff_entry[2])

        expanded_sorted_lines = [
            line
            for file, lines in sorted(
                new_problem_lines_by_file.items(), key=lambda k_v: k_v[0] or ''
            )
            for line in lines
        ]

        self._write_lines_filepath(self.problem_lines_filepath, expanded_sorted_lines)

    def __str__(self):
        return self.config.name

    def _read_lines_filepath(self, filepath):
        if not os.path.exists(filepath):
            return []

        with open(filepath) as file:
            return [self.config.normalize_path(line) for line in file.readlines()]

    def _write_lines_filepath(self, filepath, problem_lines):
        os.makedirs(os.path.dirname(filepath), exist_ok=True)
        with open(filepath, 'w') as file:
            for line in problem_lines:
                print(line, file=file)



import atexit
import logging
import os
from queue import SimpleQueue
from signal import SIGTERM
from subprocess import Popen, PIPE
from threading import Thread

from .monitor_session import MonitorSession
from .settings import DEFAULT_IGNORED_DIRECTORY_NAMES
from .utils import lf

log = logging.getLogger(__name__)


class Lintmon:
    @property
    def badges(self):
        return ''.join(sess.badge for sess in self.sessions)

    def __init__(self, config):
        self.config = config
        self.sessions = []
        self.fswatch_proc = None
        self.files_queue = None

    def load_latest_sessions(self):
        new_sessions = self.new_sessions([])
        for sess in new_sessions:
            sess.skip()

        self.sessions = new_sessions

    def start_fswatch(self):
        excluded_dir_paths_options = [
            arg
            for dirpattern in (f'{dname}/' for dname in DEFAULT_IGNORED_DIRECTORY_NAMES)
            for arg in ['--exclude', dirpattern]
        ]

        args = ['fswatch', '--extended', *excluded_dir_paths_options, os.curdir]
        log.info('Starting fswatch')
        log.debug('Args: %s', args)
        self.files_queue = SimpleQueue()
        self.fswatch = Popen(args, stdout=PIPE, stderr=PIPE, encoding='utf8',)
        atexit.register(self.fswatch.terminate)

        # non-daemon thread doesn't need to be joined or terminated: will exit when main thread
        # exits
        self.reader = Thread(target=self.reader_main)
        self.reader.start()

    def stop_fswatch(self):
        log.debug('Stopping fswatch')
        self.fswatch.terminate()

    def update_sessions(self, files):
        new_sessions = self.new_sessions(files)
        for mp in new_sessions:
            mp.start()
        for mp in new_sessions:
            mp.join()
            mp.save()

        self.sessions = new_sessions

    def get_next_files(self):
        # block for the first line
        lines = [self.files_queue.get()]
        while not self.files_queue.empty():
            lines.append(self.files_queue.get())

        return lines

    def new_sessions(self, files):
        return [
            MonitorSession(monitor_config, lf(monitor_config.includes_file, files))
            for monitor_config in self.config.monitors.values()
        ]

    def reader_main(self):
        for line in self.fswatch.stdout:
            line = self.config.normalize_path(line)
            log.debug('Got another path: %s', line)
            self.files_queue.put(line)

        log.debug('fswatch stdout closed, killing self')
        os.kill(os.getpid(), SIGTERM)

    def run(self):
        # Doesn't return
        self.start_fswatch()

        try:
            while True:
                next_files = self.get_next_files()
                assert len(next_files)
                log.info('Files changed:')
                for file in next_files:
                    log.info(f'  {file}')
                self.update_sessions(next_files)
        except BaseException as exc:
            log.debug('Exiting due to exception %s', exc)
            self.stop_fswatch()
            raise



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lintmon']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0', 'psutil>=5.8.0']

entry_points = \
{'console_scripts': ['lintmon-run-all = lintmon:run_all',
                     'lintmon-start = lintmon:start',
                     'lintmon-status = lintmon:status',
                     'lintmon-status-prompt = lintmon:status_prompt',
                     'lintmon-stop = lintmon:stop',
                     'lintmond = lintmon:lintmond']}

setup_kwargs = {
    'name': 'lintmon',
    'version': '0.1.1',
    'description': 'Monitor project files in the background for lint issues and flag issues in your prompt',
    'long_description': '# lintmon\n\nA tool for monitoring for lint errors. Provides a configurable way to specify checks to run on code in a given directory when it changes on disk, run those checks in the background and provide a fast way to display the outcome of the latest checks with a concise format suitable for embedding in your prompt.\n\nIt is designed for lint errors because these are quick to run and affect single files at a time, so in general they can be checked in the background with relatively low system impact, but in principle you could use it to trigger any sort of validation check, such as running unit tests.\n\n## Motivation\n\nThe two main ways of avoiding pushing lint errors to a repo it seemed to me were:\n\n1. lint on save in your IDE\n2. add git hooks on push to check for / fix lint errors\n\nI didn\'t like either of those, because in both cases they happen synchronously while you are doing something and can cause disruptions. Also I push from the command line, so what I wanted was a warning at the point that I was going to push that was pre-calculated about errors.\n\nThis is that.\n\n## Configure\n\nInstall lintmon using `pip` in your virtualenv (or wherever you install packages for use with your codebase):\n\n    pip install lintmon\n\nAdd a file like the lintmon.yaml file in this codebase to the root directory of your codebase. It tells dirmmon what checks to do on files. It aims to be very configurable.\n\nAdd the following to your prompt in your `.bashrc` or `.zshrc`:\n\n    PS1=\'$([[ -e lintmon.yaml ]] && which lintmon-status-prompt >/dev/null && lintmon-status-prompt)\'$PS1\n\nThat\'s it! `lintmon-status-prompt` will start `lintmond` automatically in the background when it is run, and from now on you should get a "badge" in your prompt when there are lint errors in your directory, which will be updated when you modify files.\n\n## Commands\n\n### `lintmon-status`\n\nOutput status of lintmon along with all current errors from the linters.\n\n### `lintmon-stop`, `lintmon-start`\n\nTell lintmon to stop (and not restart automatically) or start again. Mostly useful for debugging lintmon. This will display an ` S ` badge in your prompt to tell you it is stopped.\n\n### `lintmon-run-all`\n\nRun all linters on all appropriate files in your project, thus "hydrating" lintmon\'s state if it hasn\'t been running for a while and changes have been made.\n\n### `lintmond`\n\nRun the daemon in the shell (again mainly useful for debugging).\n\n\n## Directory structure\n\nlintmon adds a `.lintmon` directory to your project directory where it stores all its state about what current errors there are, lintmon\'s pid etc. You will probably want to add this to your .gitignore.\n',
    'author': 'David Park',
    'author_email': 'david@greenparksoftware.co.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/daphtdazz/lintmon',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
